@@ -11,7 +11,7 @@ from queue import Queue
 import traceback
 import inspect
 from redflagbpm.Services import Service, DocumentService, Context, ResourceService, ExecutionService, RuntimeService, \
-    ReportService
+    ReportService, Logger, AgentService
 
 
 class ServiceError(Exception):
@@ -27,6 +27,8 @@ class BPMService:
     documentService: DocumentService
     resourceService: ResourceService
     runtimeService: RuntimeService
+    agentService: AgentService
+    logger: Logger
 
     __eb_calls: EventBus
     __eb_handlers: EventBus
@@ -62,10 +64,18 @@ class BPMService:
         self.resourceService = ResourceService(self)
         self.runtimeService = RuntimeService(self)
         self.reportService = ReportService(self)
+        self.agentService = AgentService(self)
+        self.logger = Logger(self)
 
         self.connect()
         atexit.register(self.close)
 
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        self.close()
+
     def __setAddress(self, address: Any = None):
         """
         Este método se utiliza durante la conexión con el backend de la BPM.
@@ -127,6 +137,9 @@ class BPMService:
         if do_quit:
             quit()
 
+    def publish(self, address, body=None, headers=None):
+        self.__eb_calls.send(address=address, headers=headers, body=body)
+
     def call(self, address, body=None, headers=None):
         ret = self.send(address, headers, body)
         if 'body' in ret:
@@ -233,7 +246,7 @@ class BPMService:
         self.register_handler(address, handler)
         self.start()
 
-    def exec(self, script: str, lang: str = 'javascript', context: dict = None, address: object = None) -> object:
+    def exec(self, script: str, lang: str = 'juel', context: dict = None, address: object = None) -> object:
         addr = self.__getAddress(address)
         ret = Queue()
         self.__eb_calls.send(address=addr, body={
