@@ -1,5 +1,5 @@
-from typing import List, Any
 import os
+from typing import List, Any, Literal
 from urllib.parse import urlparse
 
 
@@ -68,7 +68,7 @@ class Service:
         """
         Emite una notificación a un usuario logueado en la BPM.
 
-        :param user: el nombre del usuario
+        :param group: el nombre del groupo
         :param title: el título de la notificación
         :param description: la descripción de la notificación
         :param target: destino de la notificación (cuando hace clic se puede redirigir a otra sección de la BPM o url)
@@ -102,17 +102,65 @@ class Service:
         return self.bpm.call("EBHBPMService.text", body={
             "name": name})
 
+    def checkCron(self, cron: str):
+        return self.bpm.call("EBHBPMService.checkCron", body={
+            "cron": cron})
+
     def code(self, name: str):
         return self.bpm.call("EBHBPMService.code", body={
             "name": name})
 
+    def code_properties(self, name: str):
+        return self.bpm.call("EBHBPMService.codeProperties", body={
+            "name": name})
+
     def list(self, name: str, the_filter: dict = None):
         return self.bpm.call("EBHBPMService.list", body={
             "name": name,
             "filter": the_filter})
 
+    def tokenize(self, user: str, payload: dict = None):
+        return self.bpm.call("EBHBPMService.tokenize", body={
+            "user": user, "payload": payload})
+
+
+class Logger:
+    def __init__(self, bpm):
+        self.bpm = bpm
+
+    def log(self, logger: str, message: str,
+            level: Literal["info", "warn", "error", "debug", "trace"] = "info"):
+        """
+        Logs a message with the specified logger and level.
+
+        :param message: The message to log
+        :param logger: The logger to use (default is "bpm.general")
+        :param level: The log level (default is "info"). One of: "info", "warn", "error", "debug", "trace"
+        """
+        return self.bpm.call("EBHBPMService.log", body={
+            "message": message, "logger": logger, "level": level})
+
+    def info(self, logger: str, message: str):
+        return self.log(logger, message, "info")
+
+    def warn(self, logger: str, message: str):
+        return self.log(logger, message, "warn")
+
+    def error(self, logger: str, message: str):
+        return self.log(logger, message, "error")
+
+    def debug(self, logger: str, message: str):
+        return self.log(logger, message, "debug")
+
+    def trace(self, logger: str, message: str):
+        return self.log(logger, message, "trace")
+
 
 class ContextInput(object):
+
+    def __init__(self):
+        self.context = None
+
     def __getitem__(self, item):
         return self.context.getInputValue(item)
 
@@ -121,6 +169,11 @@ class ContextInput(object):
 
 
 class ContextJsonValue(object):
+
+    def __init__(self):
+        self.context = None
+        self.item = None
+
     def __getitem__(self, item):
         return self.context.getJsonValue(self.item, item)
 
@@ -201,7 +254,7 @@ class Context(object):
 
     def setValue(self, variable: str, value: Any, address=None):
         try:
-            self.bpm.exec(script="_____context_____.setAttribute('" + variable + "',_____variable_____,100)",
+            self.bpm.exec(script=f"${{_____context_____.setAttribute('{variable}',_____variable_____,100)}}",
                           context={"_____variable_____": value},
                           address=address)
         except ValueError as x:
@@ -212,22 +265,14 @@ class Context(object):
 
     def setJsonValue(self, variable: str, key: str, value: Any, address=None):
         try:
-            self.bpm.exec(script=variable + ('.put("%s",_____variable_____)' % key),
-                          context={"_____variable_____": value},
-                          address=address)
-        except ValueError as x:
-            if str(x) == 'No address supplied':
-                if variable not in self.fakeContext.keys():
-                    self.fakeContext[variable] = {}
-                self.fakeContext[variable][key] = value
+            if value is None:
+                self.bpm.exec(script=f"${{{variable}.putNull('{key}')}}",
+                              context={},
+                              address=address)
             else:
-                raise x
-
-    def setDictValue(self, variable: str, key: str, value: Any, address=None):
-        try:
-            self.bpm.exec(script=variable + ('["%s"]=_____variable_____' % key),
-                          context={"_____variable_____": value},
-                          address=address)
+                self.bpm.exec(script=f"${{{variable}.put('{key}',_____variable_____)}}",
+                              context={"_____variable_____": value},
+                              address=address)
         except ValueError as x:
             if str(x) == 'No address supplied':
                 if variable not in self.fakeContext.keys():
@@ -240,15 +285,16 @@ class Context(object):
         self.setJsonValue("input", key, value, address)
 
     def getValue(self, variable: str, address=None):
-        return self.bpm.exec(script=variable, address=address)
+        return self.bpm.exec(script=f"${{{variable}}}", address=address)
+
+    def get(self, variable: str, default: Any = None):
+        try:
+            return self.__getitem__(variable)
+        except KeyError:
+            return default
 
     def getJsonValue(self, variable: str, key: str, address=None):
-        return self.bpm.exec(script=variable + ('.getValue("%s")' % key),
-                             address=address)
-
-    def getDictValue(self, variable: str, key: str, address=None):
-        return self.bpm.exec(script=variable + ('["%s"]' % key),
-                             address=address)
+        return self.bpm.exec(script=f"${{{variable}.getValue(\"{key}\")}}", address=address)
 
     def getInputValue(self, key: str, address=None):
         return self.getJsonValue("input", key, address)
@@ -438,6 +484,66 @@ class RuntimeService:
                       body={"executionId": executionId, "messageName": messageName,
                             "processVariables": processVariables})
 
+    def setTaskOwner(self, priority: int, taskId: str = None, processInstanceId: str = None,
+                     processBusinessKey: str = None, taskName: str = None):
+        """
+        Set the owner of a task.
+
+        :param priority: The priority to set for the task
+        :param taskId: The ID of the task
+        :param processInstanceId: The ID of the process instance (used if taskId is not provided)
+        :param processBusinessKey: The business key of the process instance (used if taskId is not provided)
+        :param taskName: The name of the task (used if taskId is not provided)
+        :return:
+        """
+        self.bpm.call("EBHBPMRuntime.setTaskPriority", body={
+            "priority": priority,
+            "taskId": taskId,
+            "processInstanceId": processInstanceId,
+            "processBusinessKey": processBusinessKey,
+            "taskName": taskName
+        })
+
+    def setTaskOwner(self, userId: str, taskId: str = None, processInstanceId: str = None,
+                        processBusinessKey: str = None, taskName: str = None):
+        """
+        Set the owner of a task.
+
+        :param userId: The ID of the user to assign the task to
+        :param taskId: The ID of the task
+        :param processInstanceId: The ID of the process instance (used if taskId is not provided)
+        :param processBusinessKey: The business key of the process instance (used if taskId is not provided)
+        :param taskName: The name of the task (used if taskId is not provided)
+        :return:
+        """
+        self.bpm.call("EBHBPMRuntime.setTaskOwner", body={
+            "userId": userId,
+            "taskId": taskId,
+            "processInstanceId": processInstanceId,
+            "processBusinessKey": processBusinessKey,
+            "taskName": taskName
+        })
+
+    def setTaskAssignee(self, userId: str, taskId: str = None, processInstanceId: str = None,
+                        processBusinessKey: str = None, taskName: str = None):
+        """
+        Set the assignee of a task.
+
+        :param userId: The ID of the user to assign the task to
+        :param taskId: The ID of the task
+        :param processInstanceId: The ID of the process instance (used if taskId is not provided)
+        :param processBusinessKey: The business key of the process instance (used if taskId is not provided)
+        :param taskName: The name of the task (used if taskId is not provided)
+        """
+        self.bpm.call("EBHBPMRuntime.setTaskAssignee", body={
+            "userId": userId,
+            "taskId": taskId,
+            "processInstanceId": processInstanceId,
+            "processBusinessKey": processBusinessKey,
+            "taskName": taskName
+        })
+
+
 
 class ExecutionService:
     def __init__(self, bpm):
@@ -450,13 +556,21 @@ class ExecutionService:
 
     def updateBusinessKey(self, businessKey: str):
         process_instance_id = self.bpm.context.getValue('execution.getProcessInstanceId()')
-        self.bpm.call("EBHBPMRuntime.updateBusinessKey", body={
-            "processInstanceId": process_instance_id, "businessKey": businessKey})
+        try:
+            self.bpm.exec(script="${runtimeService.updateBusinessKey(________pi,________bk)}",
+                          context={"________pi": process_instance_id, "________bk": businessKey})
+        except ValueError as x:
+            self.bpm.call("EBHBPMRuntime.updateBusinessKey", body={
+                "processInstanceId": process_instance_id, "businessKey": businessKey})
 
     def setProcessInstanceName(self, name: str):
         process_instance_id = self.bpm.context.getValue('execution.getProcessInstanceId()')
-        self.bpm.call("EBHBPMRuntime.setProcessInstanceName",
-                      body={"processInstanceId": process_instance_id, "name": name})
+        try:
+            self.bpm.exec(script="${runtimeService.setProcessInstanceName(________pi,________name)}",
+                          context={"________pi": process_instance_id, "________name": name})
+        except ValueError as x:
+            self.bpm.call("EBHBPMRuntime.setProcessInstanceName",
+                          body={"processInstanceId": process_instance_id, "name": name})
 
     def getVariables(self):
         process_instance_id = self.bpm.context.getValue('execution.getProcessInstanceId()')
@@ -474,16 +588,17 @@ class ExecutionService:
 
     def setVariable(self, variableName: str, value: Any):
         try:
-            self.bpm.exec(script="execution.setVariable(_____variable_____,_____value_____)",
-                          context={"_____variable_____": variableName, "_____value_____":value})
+            self.bpm.exec(script="${execution.setVariable(_____variable_____,jsonHelper.toFlowable(_____value_____))}",
+                          context={"_____variable_____": variableName, "_____value_____": value})
         except ValueError as x:
             process_instance_id = self.bpm.context.getValue('execution.getProcessInstanceId()')
             return self.bpm.call("EBHBPMRuntime.setVariable",
-                                 body={"executionId": process_instance_id, "variableName": variableName, "value": value})
+                                 body={"executionId": process_instance_id, "variableName": variableName,
+                                       "value": value})
 
     def removeVariable(self, variableName: str):
         try:
-            self.bpm.exec(script="execution.removeVariable(_____variable_____)",
+            self.bpm.exec(script="${execution.removeVariable(_____variable_____)}",
                           context={"_____variable_____": variableName})
         except ValueError as x:
             process_instance_id = self.bpm.context.getValue('execution.getProcessInstanceId()')
@@ -493,3 +608,27 @@ class ExecutionService:
     def getProcessInstanceId(self):
         process_instance_id = self.bpm.context.getValue('execution.getProcessInstanceId()')
         return process_instance_id
+
+
+class AgentService:
+    def __init__(self, bpm):
+        self.bpm = bpm
+
+    def startChat(self, agent: str, userId: str, handler: callable, body=None):
+        if body is None:
+            body = {}
+        address: str = self.bpm.call("EBHBPMAgentService.startChat",
+                                     body={"agent": agent, "userId": userId, "body": body})
+        self.bpm.register_handler(address, handler)
+        return address
+
+    def chat(self, chatId: str, message: str, body=None) -> bool:
+        if body is None:
+            body = {}
+        return self.bpm.call("EBHBPMAgentService.chat", body={"chatId": chatId, "message": message, "body": body})
+
+    def endChat(self, chatId: str, body=None) -> bool:
+        if body is None:
+            body = {}
+        self.bpm.unregister_handler(chatId)
+        return self.bpm.call("EBHBPMAgentService.endChat", body={"chatId": chatId, "body": body})
