@@ -6,7 +6,7 @@ class MyFakeDict:
 #Real dict
 class MyDict(dict):
     def test(self):
-        return 'culo';
+        return '1234';
 
 
 print("------------- dict REAL --------------")
